#define _POSIX_C_SOURCE 200112L /* needed for struct addrinfo */

#include "db.h"
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <errno.h>

typedef struct {
    char *tmp_dir;
    char *data_dir;
    char *gallery_dl_conf;
    char *yt_dlp_conf;
} Config;

int config_parser(Config *config) {
    char *config_location = malloc(10000);
    char *user_name = getenv("USER");
    if (user_name == NULL) {
        fprintf(stderr, "Couldn't read env variable USER\n");
        return 1;
    }
    sprintf(config_location, "/home/%s/.config/HDB.conf", user_name);
    
    FILE *fptr = fopen(config_location, "r");
    if (fptr == NULL) {
        fprintf(stderr, "Failed to open the config %s\n", config_location);
        return 1;
    }

    char *line = malloc(10000);
    char *left = malloc(10000);
    char *right = malloc(10000);

    while(fgets(line, 10000, fptr) != NULL) {
        line[strcspn(line, "\n#")] = '\0';
        if (strcmp(line, "") == 0) {
            continue;
        }
        if (
            strchr(line, '=') == NULL
            || line[strcspn(line, "=") + 1] == ' '
            || line[strcspn(line, "=") - 1] == ' '
        ) {
            fprintf(stderr, "Config line doesn't contain a valid assignment. line = '%s'\n", line);
            continue;
        }
        /* TODO: support env variables */
        if (strchr(line, '$') != NULL) {
            fprintf(stderr, "Config doesn't support env variables. line = '%s'\n", line);
            continue;
        }
        
        left = strtok(line, "=");
        right = strtok(NULL, "\0");
        if (strcmp(left, "data_dir") == 0) {
            config->data_dir = malloc(10000);
            memcpy(config->data_dir, right, strlen(right));
        } else if (strcmp(left, "gallery-dl_conf") == 0) {
            config->gallery_dl_conf = malloc(10000);
            memcpy(config->gallery_dl_conf, right, strlen(right));
        } else if (strcmp(left, "yt-dlp_conf") == 0) {
            config->yt_dlp_conf = malloc(10000);
            memcpy(config->yt_dlp_conf, right, strlen(right));
        } else {
            fprintf(stderr, "Unknown config option. left = '%s'\n", left);
        }
    }
    fclose(fptr);
    free(line);
    free(config_location);
    return 0;
}

int download(char *line, Config config) {
    int ret;
    char *url = malloc(strlen(line));
    char *artist_name = malloc(strlen(line));
    char *website_name = malloc(strlen(line));
    sscanf(line, "%s %s %s", url, artist_name, website_name);
    char *args[7]; 
    char *tmp = malloc(10000);
    if (strstr(website_name, "Iwara") != NULL) {
        sprintf(tmp, "%s/%s/%s/%%(id)s.%%(ext)s", config.data_dir, artist_name, website_name);
        args[0] = "yt-dlp";
        args[1] = "--config-locations";
        args[2] = config.yt_dlp_conf;
        args[3] = "-o";
        args[4] = tmp;
        args[5] = url;
    } else {
        sprintf(tmp, "%s/%s/%s", config.data_dir, artist_name, website_name);
        args[0] = "gallery-dl";
        args[1] = "--config";
        args[2] = config.gallery_dl_conf;
        args[3] = "--directory";
        args[4] = tmp;
        args[5] = url;
    }
    args[6] = NULL;
    
    printf("Starting download using %s\n", args[0]);
    /* Keep for later debugging or logging
    size_t j;
    for (j = 0; args[j] != NULL; j++) {
        printf("args[%d] = %s\n", j, args[j]);
    }
    */

    pid_t pid = fork();

    if (pid == -1) {
        printf("Couldn't fork'\n");
        return 1;
    } else if (pid > 0) {
        waitpid(pid, &ret, 0);
    } else {
        execvp(args[0], args);
    }

    if (ret != 0) {
        fprintf(stderr, "Command exited with non-zero code. code = %d\n", ret);
    }

    free(url);
    free(artist_name);
    free(website_name);
    free(tmp);
    return 0;
}

int main(int argc, char **argv) {
    int ret;
    Config config = { 0 };
    config.tmp_dir = "/tmp/HDB";
    
    ret = config_parser(&config);
    if (ret != 0) {
        fprintf(stderr, "Failed to parse config\n");
        return 1;
    }

    if(mkdir(config.tmp_dir, 0777) && errno != EEXIST) {
        fprintf(stderr, "'%s' %s\n", config.tmp_dir, strerror(errno));
        return 1;
    }

    if (argc < 2) {
        fprintf(stderr, "Incorrect amount of arguments\n");
        return 1;
    }

    ret = open("HDB.db");
    if (ret != 0) {
        return 1;
    }

    if (strcmp(argv[1], "add") == 0) {
        char *db_row = malloc(strlen(argv[2]) * 2);
        memset(db_row, 0, strlen(argv[2]));

        if (argc == 3) {
            ret = db_add(db_row, argv[2]);
            if (ret != 0) {
                db_close();
                return 1;
            } 
        } else if (argc == 4) {
            ret = db_add_custom_name(db_row, argv[2], argv[3]);
            if (ret != 0) {
                db_close();
                return 1;
            }
        } else {
            db_close();
            fprintf(stderr, "Incorrect amount of arguments\n");
            return 1;
        }
        printf("Successfully added to DB\n");
        ret = download(db_row, config);
        free(db_row);
        if (ret != 0) {
            printf("Failed download\n");
            db_close();
            return 1;
        }
        printf("Succesfully downladed\n");
    } else if (strcmp(argv[1], "download_website") == 0) {
        if (argc != 3) {
            fprintf(stderr, "Incorrect amount of arguments\n");
            db_close();
            return 1;
        }
        char *file_name = malloc(10000);
        sprintf(file_name, "%s/%s_generated.txt", config.tmp_dir, argv[2]);
        FILE *fptr = fopen(file_name, "w+");
        if (fptr == NULL) {
            fprintf(stderr, "Couldn't open file '%s'\n", file_name);
            db_close();
            return 1;
        }

        ret = db_generate_file(fptr, argv[2]);
        fclose(fptr);
        if (ret != 0) {
            db_close();
            return 1;
        }

        fptr = fopen(file_name, "r");
        if (fptr == NULL) {
            fprintf(stderr, "Couldn't open file '%s'\n", file_name);
            db_close();
            return 1;
        }
        char *line = malloc(10000);
        printf("Ready to download\n");
        while(fgets(line, 10000, fptr) != NULL) {
            ret = download(line, config);
            if (ret != 0) {
                fclose(fptr);
                db_close();
                return 1;
            }
        }
        
        free(line);
        fclose(fptr);
        free(file_name);
    } else {
        fprintf(stderr, "Invalid argument\n");
        db_close();
        return 1;
    }

    db_close();
    return 0;
}
